<?php
/**
* 2007-2015 PrestaShop
*
* NOTICE OF LICENSE
*
* This source file is subject to the Academic Free License (AFL 3.0)
* that is bundled with this package in the file LICENSE.txt.
* It is also available through the world-wide-web at this URL:
* http://opensource.org/licenses/afl-3.0.php
* If you did not receive a copy of the license and are unable to
* obtain it through the world-wide-web, please send an email
* to license@prestashop.com so we can send you a copy immediately.
*
* DISCLAIMER
*
* Do not edit or add to this file if you wish to upgrade PrestaShop to newer
* versions in the future. If you wish to customize PrestaShop for your
* needs please refer to http://www.prestashop.com for more information.
*
*  @author    PrestaShop SA <contact@prestashop.com>
*  @copyright 2007-2015 PrestaShop SA
*  @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
*  International Registered Trademark & Property of PrestaShop SA
*/

class OfdfermaApi
{
    const API_URL   = 'https://ferma.ofd.ru/api/kkt/cloud';
    const AUTH_URL  = 'https://ferma.ofd.ru/api/Authorization/CreateAuthToken';
    const CHECK_URL = 'https://check.ofd.ru/rec/';
    
    private $debug = false;
    
    private $settings = array();
    
    public static $reference_vat = array(
        array(
            'id'   => 'Vat0',
            'name' => 'Без НДС',
        ),
        array(
            'id'   => 'Vat10',
            'name' => '10%',
        ),
        array(
            'id'   => 'Vat18',
            'name' => '18%',
        ),
        array(
            'id'   => 'Vat20',
            'name' => '20%',
        ),
        array(
            'id'   => 'CalculatedVat10110',
            'name' => '10/110',
        ),
        array(
            'id'   => 'CalculatedVat18118',
            'name' => '18/118',
        ),
        array(
            'id'   => 'CalculatedVat20120',
            'name' => '20/120',
        ),
    );
    
    public static $reference_nalog = array(
        array(
            'id'    => 'Common',
            'name'  => 'Общая',
        ),
        array(
            'id'    => 'SimpleIn',
            'name'  => 'Упрощенная',
        ),
        array(
            'id'    => 'SimpleInOut',
            'name'  => 'Упрощенная (доходы минус расходы)',
        ),
        array(
            'id'    => 'Unified',
            'name'  => 'Единый налог на вмененный доход',
        ),
        array(
            'id'    => 'UnifiedAgricultural',
            'name'  => 'Единый сельскохозяйственный налог',
        ),
        array(
            'id'    => 'Patent',
            'name'  => 'Патент',
        ),
    );
    
    private static $instance;
    
    public static function getInstance()
    {
        if (null === static::$instance) {
            static::$instance = new static();
        }

        return static::$instance;
    }
    
    function __construct() {
        
        $this->settings = array(
            'login'         => Configuration::get('ofdferma_login'),
            'password'      => Configuration::get('ofdferma_password'),
            'osn'           => Configuration::get('ofdferma_nalog'),
            'inn'           => Configuration::get('ofdferma_inn'),
            'nds'           => Configuration::get('ofdferma_vat'),
            'collapse'      => Configuration::get('ofdferma_collapse'),
            'collapse_name' => Configuration::get('ofdferma_collapse_name'),
            'status'        => Configuration::get('ofdferma_status'),
            'email'         => Configuration::get('ofdferma_email'),
            
            'token'         => Configuration::get('ofdferma_token'),
            'token_exp_date'=> Configuration::get('ofdferma_token_exp_date')
        );
    }
    
    public function setDebug($val) {
        $this->debug = $val;
    }
    
    private function check_settings() {
        $ofd_login      = $this->settings['login'];
        $ofd_pass       = $this->settings['password'];
        $ofd_osn        = $this->settings['osn'];
        $ofd_inn        = $this->settings['inn'];
        $ofd_nds        = $this->settings['nds'];

        return (
            $ofd_login &&
            $ofd_pass &&
            $ofd_osn &&
            $ofd_inn &&
            $ofd_nds
        );
    }
    
    public function getTextType($type) {
        $res = '';

        switch ($type) {
            case 'Income':
                $res = 'Приход';
                break;

            case 'IncomeReturn':
                $res = 'Возврат';
                break;
        }

        return $res;
    }
    
    private function getErrorByCode($code_id, $data) {
        $msg = '';

        SWITCH ($code_id) {
			//Неверная авторизация
			case 2:
				$msg = 'Не удалось обновить статус чека. Авторизация не возможна, неверные учетные данные.';
			break;
			
            case 1019:
                $msg = "Идентификатор счета '{$data['Request']['InvoiceId']}' уже существует в ОФД";
                break;
        }

        return $msg;
    }
    
    private function formatFloat( $int, $count = 2){
        return	round( (float)$int , $count);
    }
    
    private function customErrorHandler($errno, $errstr, $errfile, $errline, array $errcontext) {
        $this->log_msg($errstr);

        if (0 === error_reporting()) {
            return false;
        }
        throw new Exception($errstr, 0, $errno, $errfile, $errline);
    }
    
    public static function getCheckLink($row){
        $check_link = self::CHECK_URL . "{$row['inn']}/{$row['RNM']}/{$row['FN']}/".ltrim($row['FDN'],0)."/".$row['FPD'];
        
        return $check_link;
    }
    
    public static function get_product_vat($tovar_id, $default = ''){
        $nds = Db::getInstance(_PS_USE_SQL_SLAVE_)->getValue('SELECT nds_id FROM `' . _DB_PREFIX_ . 'ofdferma_tovars` WHERE `tovar_id`="' . (int)$tovar_id . '"');
        return $nds ? $nds : $default;
    }
    
    private function get_receipt($order_id, $type){
        $val = Db::getInstance(_PS_USE_SQL_SLAVE_)->getValue('SELECT id_ofdferma_receipt FROM `' . _DB_PREFIX_ . 'ofdferma_receipt` WHERE `order_id` = "' . (int)$order_id . '" && `type` = "' .addslashes($type). '"');
        return (bool)$val;
    }
    
    public function log_msg($mess, $status = false, $params = array()) {
        if ( ! $status && $this->debug && $mess) {
            $txt = $mess;
            
            PrestaShopLogger::addLog($txt, 3, null, null, null, true);

            $user_mail = $this->settings['email'];
            if ($user_mail){  
                $id_lang = (int) Configuration::get('PS_LANG_DEFAULT');
                
                Mail::Send($id_lang, 'msg', 'OFD Ferma - Ошибка', array(
                        '{message}' => nl2br(date('d.m.Y H:i:s') .' : '. $txt),
                    ), 
                    $user_mail,
                    null,
                    null,
                    null,
                    null,
                    null,
                    _PS_MODULE_DIR_.'ofdferma/mails/'
                );
            }
        }
        
        return array_merge(array(
            'status'    => $status,
            'mess'      => $mess
        ), (array)$params);
    } 
    
    public function set_status_order($order_id, $state){
        
        $res = array('status' => false, 'status_match' => false);
        
        //Если статус является тем при котором нужно формировать чек прихода
        if ($this->settings['status'] == $state->id){
			$this->setDebug(1);
            $res = $this->Check($order_id, 'Income');
            $this->setDebug(0);
			
			$res['status_match'] = true;
        }
        
        return $res;
    }
    
    //Обновляем токен
    private function get_token(){
        
        if ( ! $this->check_settings()) {
            return $this->log_msg('Проверьте настройки, для корректной работы должны быть заполнены: Логин, Пароль, НДС по умолчанию, ИНН, Система налогообложения');
        };
        
        if ($this->settings['token'] && ($this->settings['token_exp_date'] > (time() - 10))) {
            return $this->log_msg('', true);
        }
        
        $ofd_login      = $this->settings['login'];
        $ofd_pass       = $this->settings['password'];
        
        $ofd_auth_url   = self::AUTH_URL;

        $response = $this -> execute_request($ofd_auth_url, array(
            "Login"     => $ofd_login,
            "Password"  => $ofd_pass,
        ));
        
        if (isset($response['result'])){
            $result = $response['result'];
            
            //Обновляем токен
            $this->settings['token'] = $result['Data']['AuthToken'];
            $this->settings['token_exp_date'] = $result['Data']['ExpirationDateUtc'];
            
            foreach(array('token','token_exp_date') as $field){
                Configuration::updateGlobalValue('ofdferma_'.$field, $this->settings[$field]);
            }
            
            return $this->log_msg('', true);
        }
        
        return $response;
    }
    
    private function execute_request($url, $data = null) {  
        
        $options = array(
            "ssl" => array(
                "verify_peer" => false,
                "verify_peer_name" => false,
            ),
            'http' => array(
                'timeout' => 10,
                'ignore_errors' => true,
                'content' => json_encode($data),
                'header' => "Content-type: application/json\r\n" .
                "Accept: application/json" . "\r\n",
                "Content-Length: " . strlen(json_encode($data)) . "\r\n",
                'method' => 'POST',
            )
        );
        
        $context = stream_context_create($options);
        set_error_handler(array($this, 'customErrorHandler'));

        try {
            $result = file_get_contents($url, false, $context);
        } catch (Exception $e) {
            return $this->log_msg($e->getMessage());
        }
        
        restore_error_handler();
        $result = json_decode($result, TRUE);
        
        if (isset($result['Status']) && ($result['Status'] == 'Success')) {
            return $this->log_msg('', true, array('result' => $result));
        } else if (isset($result['Status']) && ($result['Status'] == 'Failed')) {
            $msg = $result['Error']['Message'];
            if ($result['Error']['Code']) {
                if ($val = $this->getErrorByCode($result['Error']['Code'], $data)) {
                    $msg = $val;
                }
            }
            
            return $this->log_msg($msg, false, $result);
        } else {
            return $this->log_msg('some error', false, $result);
        }
    }
    
    /**
    * Отправка чека в ОФД Ferma
    * @param array $data данные заказа
    * @param string $operation операция [sale|sale_refund]
    * @param bool $json передача упакованных внешних данных в JSON формате
    */
    public function Check($order_id, $type = 'Income') {
       
        $res = $this->get_token();
        if ( ! $res['status']) {
            return $res;
        }
       
        $order = new Order($order_id);
        $order_items = $order->getOrderDetailList();
        //$order->getProductsDetail() 
        
        $temptype = $this -> getTextType($type);
        
        // Данные из заказа
        if ( ! $order_items) {
            return $this->log_msg("В заказе #{$order_id} нет товаров");
        }
        
        if ($this -> get_receipt($order_id, $type)){
            return $this->log_msg("Чек {$temptype} для заказа {$order_id} уже существует в модуле");
        }
        
        $order_customer = $order->getCustomer();
        
        //Формируем данные для запроса
        $prefix     = 'o';
        
        $ofd_osn    = $this->settings['osn'];
        $ofd_inn    = $this->settings['inn'];
        $ofd_nds    = $this->settings['nds'];
        
        //свертка
        $ofd_collapse       = $this->settings['collapse'];
        $ofd_collapse_name  = $this->settings['collapse_name'];
        
        $customer_phone = str_replace(array('(', ')', '-', ' '), '', '');
        $customer_email = $order_customer->email;     

        //Формируем данные
        $data = array();
        $data['Request']['Inn'] = $ofd_inn;
        $data['Request']['Type'] = $type;
        $data['Request']['InvoiceId'] = $prefix . $order_id . '-' . $type;
        $data['Request']['LocalDate'] = date('Y-m-d\TH:i:s');
        $data['Request']['CustomerReceipt'] = array(
            'TaxationSystem'    => $ofd_osn,
            'Email'             => $customer_email,
            'Phone'             => $customer_phone,
            'Items'             => array(),
        );
           
        
        //Товары
        if (is_array($order_items)) {
            foreach ($order_items as $product){
                $vat = $this->get_product_vat($product['product_id'], $ofd_nds);
                
                array_push($data['Request']['CustomerReceipt']['Items'], array(
                        'Label'     => $product['product_name'],
                        'Price'     => $this->formatFloat( $product['product_price'] ),
                        'Quantity'  => $this->formatFloat( $product['product_quantity'], 3),
                        'Amount'    => $this->formatFloat( $product['product_price'] * $product['product_quantity'] ),
                        'Vat'       => $vat,
                    )
                );
            }
        }

        // Доставка
        if (!empty((float)$order->total_shipping)) {
            array_push($data['Request']['CustomerReceipt']['Items'], array(
                    'Label'     => 'Доставка',
                    'Price'     => $this->formatFloat( $order->total_shipping ),
                    'Quantity'  => $this->formatFloat( 1, 3),
                    'Amount'    => $this->formatFloat( $order->total_shipping ),
                    'Vat'       => $ofd_nds,
                )
            );
        }
        
        
        //Если включена свертка
        if ($ofd_collapse) {
            $pos_name = $ofd_collapse_name ? $ofd_collapse_name : 'Undefined';

            $sum = 0;
            foreach ($data['Request']['CustomerReceipt']['Items'] as $item) {
                $sum += $item['Amount'];
            }
            $data['Request']['CustomerReceipt']['Items'] = array(array(
                'Label'     => $pos_name,
                'Price'     => $sum,
                'Quantity'  => 1,
                'Amount'    => $sum,
                'Vat'       => $ofd_nds,
            ));
        }
        
        
        //Отправляем запрос к ОФД
        $api_url = self::API_URL . "/receipt?AuthToken=" . $this->settings['token'];
        $response = $this -> execute_request($api_url, $data);
       
        if (isset($response['result'])){
            $result = $response['result'];
            
            // Запись лога
            Db::getInstance()->insert('ofdferma_receipt', array(
                'id_ofdferma_receipt'       => $result['Data']['ReceiptId'],
                'type'                      => $data['Request']['Type'],
                'inn'                       => $data['Request']['Inn'],
                'order_id'                  => $order_id,
                'total'                     => $order->total_paid,
                'created_at'                => gmdate('Y-m-d H:i:s')
            ), false, false, Db::INSERT);
                 
            return $this->log_msg("Чек {$temptype} для заказа {$order_id} сформирован: {$result['Data']['ReceiptId']}", true);
        }else{
            return $response;
        }
    }
   
    //Обновляем статус чеков
    public function check_update($id){
        
        $res = $this->get_token();
        if ( ! $res['status']) {
            return $res;
        }

        $results = Db::getInstance(_PS_USE_SQL_SLAVE_)->executeS('SELECT * FROM `' . _DB_PREFIX_ . 'ofdferma_receipt` WHERE `id_ofdferma_receipt` = "' .addslashes($id). '" && `status` NOT IN ("FAILED","CONFIRMED")');
        
        $this -> setDebug(1);
        foreach ($results as $result) {
            $data = array();
            $data['Request']['ReceiptId'] = $result['id_ofdferma_receipt'];
            
            //Отправляем запрос к ОФД
            $api_url = self::API_URL . "/status?AuthToken=" . $this->settings['token'];
            $response = $this -> execute_request($api_url, $data);

            if (isset($response['result'])){
                // Запись лога                
                Db::getInstance()->update('ofdferma_receipt', array(
                    'status'                    => $response['result']['Data']['StatusName'],
                    'status_message'            => $response['result']['Data']['StatusMessage'],
                    'FN'                        => $response['result']['Data']['Device']['FN'],
                    'RNM'                       => $response['result']['Data']['Device']['RNM'],
                    'FDN'                       => $response['result']['Data']['Device']['FDN'],
                    'FPD'                       => $response['result']['Data']['Device']['FPD'],
                    'updated_at'                => gmdate('Y-m-d H:i:s')
                ), "`id_ofdferma_receipt` = '". addslashes($result['id_ofdferma_receipt'])."'", 0, false, false);
            }
        }
        $this -> setDebug(0);
    }
}
